from textual import on, log
from textual.app import ComposeResult
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.screen import Screen
from textual.widgets import Footer

from elia_chat.chats_manager import ChatsManager
from elia_chat.widgets.agent_is_typing import AgentIsTyping
from elia_chat.widgets.chat import Chat
from elia_chat.widgets.chat_header import ChatHeader
from elia_chat.widgets.chat_list import ChatList
from elia_chat.widgets.chat_options import ModelSet, ModelPanel


class ChatScreen(Screen):
    """This screen doesn't correspond to a single chat. It's just the screen
    where chatting takes place. It includes the list of chats in the sidebar."""

    BINDINGS = [
        Binding("ctrl+n", action="new_chat", description="New Chat"),
        Binding(
            key="ctrl+s", action="focus('cl-option-list')", description="Focus Chats"
        ),
        Binding(key="i", action="focus('chat-input')", description="Focus Input"),
    ]

    def __init__(self):
        super().__init__()
        self.chats_manager = ChatsManager()

    def compose(self) -> ComposeResult:
        yield ChatList(id="chat-list")
        yield Chat()
        yield Footer()

    @on(Chat.UserMessageSubmitted)
    def user_message_submitted(self, event: Chat.UserMessageSubmitted) -> None:
        """Add the user message to the chat via the ChatsManager."""
        log.debug(f"In chat {event.chat_id}, user message submitted: {event.message}")
        # ChatList ordering will change, so we need to force an update...
        self.query_one(ChatList).reload_and_refresh()

    @on(Chat.AgentResponseStarted)
    def start_awaiting_response(self) -> None:
        """Prevent sending messages because the agent is typing."""
        agent_is_typing = self.query_one(AgentIsTyping)
        agent_is_typing.display = True

    @on(Chat.AgentResponseComplete)
    def agent_response_complete(self, event: Chat.AgentResponseComplete) -> None:
        """Allow the user to send messages again."""
        self.allow_input_submit = True
        agent_is_typing = self.query_one(AgentIsTyping)
        agent_is_typing.display = False

        log.debug(
            f"Agent response complete. Adding message "
            f"to chat_id {event.chat_id!r}: {event.message}"
        )

        self.chats_manager.add_message_to_chat(
            chat_id=event.chat_id, message=event.message
        )

    @on(Chat.FirstMessageSent)
    def on_first_message_sent(self, event: Chat.FirstMessageSent) -> None:
        """The first chat message was received in the current chat, so update the
        sidebar. At the time of this handler being triggered, the chat should already
        exist in the database."""
        chat_list = self.query_one(ChatList)
        chat_data = event.chat_data
        chat_list.create_chat(chat_data)

    @on(ChatList.ChatOpened)
    async def on_chat_opened(self, event: ChatList.ChatOpened) -> None:
        """Open a chat by unmounting the nodes associated with the old chat,
        and mounting the nodes associated with the new chat."""
        log.debug(f"Chat selected from chat list: {event.chat.id}")

        self.allow_input_submit = False
        chat_widget = self.query_one(Chat)
        opened_chat = self.chats_manager.get_chat(event.chat.id)

        log.debug(
            f"Retrieved chat {opened_chat.id!r} [model={opened_chat.model_name!r}] "
            f"containing {len(opened_chat.messages)} messages from database."
        )

        await chat_widget.load_chat(opened_chat)
        self.allow_input_submit = True

    @on(ModelSet.Selected)
    def update_model(self, event: ModelPanel.Selected) -> None:
        model = event.model

        try:
            conversation_header = self.query_one(ChatHeader)
        except NoMatches:
            log.error("Couldn't find ConversationHeader to update model name.")
        else:
            conversation_header.model_name = model.name

        try:
            conversation = self.query_one(Chat)
        except NoMatches:
            log.error("Couldn't find the Conversation")
        else:
            conversation.chat_data.model_name = model.name

    async def action_new_chat(self) -> None:
        chat = self.query_one(Chat)
        await chat.prepare_for_new_chat()



from __future__ import annotations

from dataclasses import dataclass

from sqlmodel import Session
from textual import log

from elia_chat.database.converters import (
    chat_dao_to_chat_data,
    chat_message_to_message_dao,
    message_dao_to_chat_message,
)
from elia_chat.database.models import ChatDao, MessageDao, engine
from elia_chat.models import ChatData, ChatMessage


@dataclass
class ChatsManager:
    @staticmethod
    def all_chats() -> list[ChatData]:
        chat_daos = ChatDao.all()
        return [chat_dao_to_chat_data(chat) for chat in chat_daos]

    @staticmethod
    def get_chat(chat_id: str) -> ChatData:
        chat_dao = ChatDao.from_id(chat_id)
        return chat_dao_to_chat_data(chat_dao)

    @staticmethod
    def get_messages(chat_id: str | int) -> list[ChatMessage]:
        with Session(engine) as session:
            try:
                chat: ChatDao | None = session.get(ChatDao, int(chat_id))
            except ValueError:
                raise RuntimeError(
                    f"Malformed chat ID {chat_id!r}. "
                    f"I couldn't convert it to an integer."
                )

            if not chat:
                raise RuntimeError(f"Chat with ID {chat_id} not found.")
            message_daos = chat.messages
            session.commit()

        # Convert MessageDao objects to ChatMessages
        chat_messages = []
        for message_dao in message_daos:
            chat_message = message_dao_to_chat_message(message_dao)
            chat_messages.append(chat_message)

        log.debug(f"Retrieved {len(chat_messages)} chats for chat {chat_id!r}")
        return chat_messages

    @staticmethod
    def create_chat(chat_data: ChatData) -> int:
        log.debug(f"Creating chat in database: {chat_data!r}")

        chat = ChatDao(model=chat_data.model_name, title="Untitled chat")

        for message in chat_data.messages:
            new_message = MessageDao(
                role=message.get("role"), content=message.get("content")
            )
            chat.messages.append(new_message)

        with Session(engine) as session:
            session.add(chat)
            session.commit()
            session.refresh(chat)

        return chat.id

    @staticmethod
    def add_message_to_chat(chat_id: str, message: ChatMessage) -> None:
        with Session(engine) as session:
            chat: ChatDao | None = session.get(ChatDao, chat_id)
            if not chat:
                raise Exception(f"Chat with ID {chat_id} not found.")
            message_dao = chat_message_to_message_dao(message)
            chat.messages.append(message_dao)
            session.add(chat)
            session.commit()



from __future__ import annotations

import asyncio
import os
import time
from dataclasses import dataclass
from typing import Any

import openai
from textual import work, log, on
from textual.app import ComposeResult
from textual.containers import VerticalScroll, Vertical, ScrollableContainer
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Input

from elia_chat.chats_manager import ChatsManager
from elia_chat.models import ChatData, ChatMessage
from elia_chat.widgets.agent_is_typing import AgentIsTyping
from elia_chat.widgets.chatbox import Chatbox
from elia_chat.widgets.chat_header import ChatHeader
from elia_chat.widgets.chat_options import (
    DEFAULT_MODEL,
    ChatOptions,
)


class Chat(Widget):
    allow_input_submit = reactive(True)
    """Used to lock the chat input while the agent is responding."""

    def __init__(self) -> None:
        super().__init__()

        self.persona_directive = os.getenv(
            "ELIA_DIRECTIVE", "You are a helpful assistant."
        )
        # The thread initially only contains the system message.
        self.chat_container: ScrollableContainer | None = None
        self.chat_options: ChatOptions | None = None
        self.chat_data = ChatData(
            id=None,
            title=None,
            create_timestamp=None,
            model_name=DEFAULT_MODEL.name,
            messages=[
                ChatMessage(
                    id=None,
                    role="system",
                    content=self.persona_directive,
                    timestamp=time.time(),
                    status=None,
                    end_turn=None,
                    weight=None,
                    metadata={},
                    recipient="all",
                )
            ],
        )

    @dataclass
    class AgentResponseStarted(Message):
        pass

    @dataclass
    class AgentResponseComplete(Message):
        chat_id: str | None
        message: ChatMessage

    @dataclass
    class FirstMessageSent(Message):
        chat_data: ChatData

    @dataclass
    class SavedChatLoaded(Message):
        chat_data: ChatData

    @dataclass
    class UserMessageSubmitted(Message):
        chat_id: str
        message: ChatMessage

    @property
    def is_empty(self) -> bool:
        """True if the conversation is empty, False otherwise."""
        return len(self.chat_data.messages) == 1  # Contains system message at first.

    def scroll_to_latest_message(self):
        if self.chat_container is not None:
            self.chat_container.refresh()
            self.chat_container.scroll_end(animate=False)

    async def new_user_message(self, content: str) -> None:
        log.debug(f"User message submitted in chat {self.chat_data.id!r}: {content!r}")
        user_message = ChatMessage(
            id=None,
            role="user",
            content=content,
            timestamp=time.time(),
            status=None,
            end_turn=None,
            weight=None,
            metadata=None,
            recipient=None,
        )
        # If the thread was empty, and now it's not, remove the ConversationOptions.
        is_first_message = self.is_empty
        if is_first_message:
            log.debug(
                f"First user message received in "
                f"conversation with model {self.chat_data.model_name!r}"
            )
            assert self.chat_options is not None
            self.chat_options.display = False

            # At this point, we should create the conversation in the database.
            # Note that we don't need to add the message here, since `create_chat`
            # already adds any of the messages already present in the chat.
            self.chat_data.id = str(ChatsManager.create_chat(chat_data=self.chat_data))

            # Add the first message to the thread AFTER creating the chat, since
            # we don't want ChatsManager.create_chat to persist the message for us.
            # The message will be persisted by us below, outwith this if block.

        self.chat_data.messages.append(user_message)
        if is_first_message:
            self.post_message(
                Chat.FirstMessageSent(
                    chat_data=self.chat_data,
                )
            )

        ChatsManager.add_message_to_chat(
            chat_id=str(self.chat_data.id), message=user_message
        )

        assert self.chat_data.id is not None
        self.post_message(
            Chat.UserMessageSubmitted(
                chat_id=str(self.chat_data.id), message=user_message
            )
        )

        assert self.chat_data.model_name is not None
        user_message_chatbox = Chatbox(user_message, self.chat_data.model_name)

        assert (
            self.chat_container is not None
        ), "Textual has mounted container at this point in the lifecycle."

        await self.chat_container.mount(user_message_chatbox)
        self.scroll_to_latest_message()
        self.post_message(self.AgentResponseStarted())
        self.stream_agent_response()

    async def clear_thread(self) -> None:
        """Remove the internal thread representing the chat, and update the DOM."""

        # We have to maintain the system message.
        self.chat_data.messages = self.chat_data.messages[:1]

        assert self.chat_container is not None

        # Clear the part of the DOM containing the chat messages.
        # Important that we make a copy before removing inside the loop!
        children = list(self.chat_container.children)
        for child in children:
            await child.remove()

    @property
    def outgoing_messages(self) -> list[dict[str, Any]]:
        return [
            {"role": message.get("role"), "content": message.get("content")}
            for message in self.chat_data.messages
        ]

    @work(exclusive=True)
    async def stream_agent_response(self) -> None:
        self.scroll_to_latest_message()
        log.debug(
            f"Creating streaming response with model {self.chat_data.model_name!r}"
        )
        streaming_response = await openai.ChatCompletion.acreate(
            model=self.chat_data.model_name,
            messages=self.outgoing_messages,
            stream=True,
        )

        # TODO - ensure any metadata available in streaming response is passed through
        message = ChatMessage(
            id=None,
            role="assistant",
            content="",
            timestamp=time.time(),
            status=None,
            end_turn=None,
            weight=None,
            metadata=None,
            recipient=None,
        )

        assert self.chat_data.model_name is not None
        response_chatbox = Chatbox(
            message=message,
            model_name=self.chat_data.model_name,
        )

        assert (
            self.chat_container is not None
        ), "Textual has mounted container at this point in the lifecycle."

        await self.chat_container.mount(response_chatbox)

        while True:
            # TODO: We need to handle RateLimitError in the worker.
            try:
                event = await streaming_response.__anext__()
                choice = event["choices"][0]
            except (StopAsyncIteration, StopIteration, IndexError):
                self.post_message(
                    self.AgentResponseComplete(
                        chat_id=self.chat_data.id,
                        message=message,
                    ),
                )
            else:
                finish_reason = choice.get("finish_reason")
                if finish_reason in {"stop", "length", "content_filter"}:
                    log.debug(
                        f"Agent response finished. Finish reason is {finish_reason!r}."
                    )
                    response_message = response_chatbox.message
                    self.post_message(
                        self.AgentResponseComplete(
                            chat_id=self.chat_data.id,
                            message=response_message,
                        )
                    )
                    return
                response_chatbox.append_chunk(event)
                scroll_y = self.chat_container.scroll_y
                max_scroll_y = self.chat_container.max_scroll_y
                if scroll_y in range(max_scroll_y - 3, max_scroll_y + 1):
                    self.chat_container.scroll_end(animate=False)
            await asyncio.sleep(0.01)

    @on(AgentResponseComplete)
    def agent_finished_responding(self, event: AgentResponseComplete) -> None:
        # Ensure the thread is updated with the message from the agent
        self.chat_data.messages.append(event.message)

    @on(Input.Submitted, "#chat-input")
    async def user_chat_message_submitted(self, event: Input.Submitted) -> None:
        if self.allow_input_submit:
            user_message = event.value
            event.input.value = ""
            await self.new_user_message(user_message)

    async def load_chat(self, chat: ChatData) -> None:
        assert self.chat_options is not None
        assert self.chat_container is not None

        # If the options display is visible, get rid of it.
        self.chat_options.display = False

        # Update the chat data
        await self.clear_thread()
        self.chat_data = chat

        assert self.chat_data.model_name is not None
        chatboxes = [
            Chatbox(chat_message, self.chat_data.model_name)
            for chat_message in chat.non_system_messages
        ]
        await self.chat_container.mount_all(chatboxes)
        self.chat_container.scroll_end(animate=False)

        chat_header = self.query_one(ChatHeader)
        chat_header.title = chat.short_preview or "Untitled Chat"
        chat_header.model_name = chat.model_name or "unknown model"
        self.post_message(Chat.SavedChatLoaded(chat))  # TODO - required?

    async def prepare_for_new_chat(self) -> None:
        await self.clear_thread()

        # Show the options to let the user configure the new chat.
        assert self.chat_options is not None

        chat_header = self.query_one(ChatHeader)
        self.chat_data.id = None
        chat_header.title = "Untitled Chat"
        chat_header.model_name = self.chat_data.model_name or "unknown model"
        self.chat_options.display = True

        log.debug(f"Prepared for new chat. Chat data = {self.chat_data}")

    def compose(self) -> ComposeResult:
        yield ChatHeader()
        with Vertical(id="chat-input-container"):
            yield Input(placeholder="[I] Enter your message here...", id="chat-input")
            yield AgentIsTyping()

        with VerticalScroll() as vertical_scroll:
            self.chat_container = vertical_scroll
            vertical_scroll.can_focus = False

        self.chat_options = ChatOptions()
        yield self.chat_options

        # TODO - check if conversation is pre-existing.
        #  If it already exists, load it here.
        #  If it's a new empty conversation, show the
        #  options for a new conversation.



from __future__ import annotations

from typing import NamedTuple

from rich.console import RenderableType
from rich.text import Text
from textual import log, on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, VerticalScroll
from textual.geometry import clamp
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Static


class GPTModel(NamedTuple):
    name: str
    icon: str
    provider: str
    product: str
    description: str
    css_class: str


DEFAULT_MODEL = GPTModel(
    name="gpt-3.5-turbo",
    icon="⚡️",
    provider="OpenAI",
    product="ChatGPT",
    description="The fastest ChatGPT model, great for most everyday tasks.",
    css_class="gpt35",
)
AVAILABLE_MODELS = [
    DEFAULT_MODEL,
    GPTModel(
        name="gpt-4",
        icon="🧠",
        provider="OpenAI",
        product="ChatGPT",
        description="The most powerful ChatGPT model, capable of "
        "complex tasks which require advanced reasoning.",
        css_class="gpt4",
    ),
]


class ModelPanel(Static):
    class Selected(Message):
        def __init__(self, model: GPTModel):
            super().__init__()
            self.model = model

    selected = reactive(False)

    def __init__(
        self,
        model: GPTModel,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.model = model

    def render(self) -> RenderableType:
        name, icon, provider, product, description, _ = self.model
        return Text.assemble(
            (f"{icon} {name}", "b"),
            "\n",
            (f"{product} by {provider} ", "italic"),
            "\n\n",
            description,
        )

    def on_click(self) -> None:
        assert self.parent is not None
        self.parent.post_message(ModelPanel.Selected(self.model))

    def watch_selected(self, value: bool) -> None:
        self.set_class(value, "selected")


class ModelSet(Horizontal, can_focus=True):
    BINDINGS = [
        Binding(key="left", action="left", description="Previous model"),
        Binding(key="right", action="right", description="Next model"),
    ]

    selected_panel_index = reactive(0)

    class Selected(Message):
        def __init__(self, model: GPTModel):
            super().__init__()
            self.model = model

    @property
    def panels(self) -> list[ModelPanel]:
        return list(self.query(ModelPanel))

    def watch_selected_panel_index(self, new_index: int) -> None:
        panel = self.panels[new_index]
        self.post_message(ModelSet.Selected(panel.model))

    @on(ModelPanel.Selected)
    def update_selection(self, event: ModelPanel.Selected) -> None:
        event.stop()
        self.focus()
        panels = self.panels
        for index, panel in enumerate(panels):
            panel.selected = panel.model == event.model
            if panel.selected:
                self.selected_panel_index = index

        log.info(f"Selected model {panels[self.selected_panel_index]}")

    def action_left(self):
        new_index = self.selected_panel_index - 1
        panels = self.panels
        self.selected_panel_index = clamp(new_index, 0, len(panels) - 1)
        for index, panel in enumerate(panels):
            panel.selected = index == self.selected_panel_index

        log.info(f"Selected model {panels[self.selected_panel_index]}")

    def action_right(self):
        new_index = self.selected_panel_index + 1
        panels = self.panels
        self.selected_panel_index = clamp(new_index, 0, len(panels) - 1)
        for index, panel in enumerate(panels):
            panel.selected = index == self.selected_panel_index

        log.info(f"Selected model {panels[self.selected_panel_index]}")


class ChatOptions(Widget):
    def compose(self) -> ComposeResult:
        with VerticalScroll(id="chat-options-container") as vertical_scroll:
            vertical_scroll.can_focus = False

            with ModelSet() as model_set:
                model_set.border_title = "Choose a language model"
                model_set.focus()
                for index, model in enumerate(AVAILABLE_MODELS):
                    model_panel = ModelPanel(
                        model, id=model.name, classes=model.css_class
                    )
                    if index == 0:
                        model_panel.selected = True

                    yield model_panel



import json
from datetime import datetime
from pathlib import Path

from sqlalchemy.orm import Session

from elia_chat.database.models import MessageDao, ChatDao, engine


def import_chatgpt_data(file: Path) -> None:
    # Load the JSON data
    with open(file, "r") as f:
        data = json.load(f)

    # Create a new session
    session = Session(engine)

    # Iterate over each chat in the data
    for chat_data in data:
        # Create a new ChatDao instance and add it to the session
        chat = ChatDao(
            model=None, started_at=datetime.fromtimestamp(chat_data["create_time"])
        )
        session.add(chat)
        session.commit()  # Make sure to commit so that chat.id is assigned

        for message_id, message_data in chat_data["mapping"].items():
            message_info = message_data.get("message")
            if message_info:
                metadata = message_info.get("metadata", {})
                if metadata:
                    model = metadata.get("model_slug")
                    if model is not None:
                        chat.model = model
                        session.add(chat)
                        session.commit()

                message = MessageDao(
                    chat_id=chat.id,
                    role=message_info["author"]["role"],
                    content=message_info["content"]["parts"][0],
                    timestamp=datetime.fromtimestamp(message_info["create_time"]),
                    status=message_info.get("status"),
                    end_turn=message_info.get("end_turn"),
                    weight=message_info.get("weight"),
                    meta=metadata,
                    recipient=message_info.get("recipient"),
                )
                session.add(message)

        session.commit()

    # Close the session when you're done
    session.close()


if __name__ == "__main__":
    path = Path("resources/conversations.json")
    import_chatgpt_data(path)



import os
from pathlib import Path

import openai
from textual.app import App

from elia_chat.screens.chat_screen import ChatScreen


class Elia(App):
    CSS_PATH = Path(__file__).parent / "elia.scss"

    def __init__(self):
        super().__init__()
        openai.api_key = os.getenv("OPENAI_API_KEY")

    def on_mount(self) -> None:
        self.push_screen(ChatScreen())


app = Elia()

if __name__ == "__main__":
    app.run()



from __future__ import annotations

import time
from typing import Any

from rich.console import RenderableType
from rich.markdown import Markdown
from textual.binding import Binding
from textual.geometry import Size
from textual.widget import Widget

from elia_chat.models import ChatMessage
from elia_chat.screens.message_info_modal import MessageInfo
from elia_chat.time_display import format_timestamp


class Chatbox(Widget, can_focus=True):
    BINDINGS = [Binding(key="d", action="details", description="Message details")]

    def __init__(
        self,
        message: ChatMessage,
        model_name: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.message = message
        self.model_name = model_name
        timestamp = format_timestamp(message.get("timestamp", 0) or 0)
        self.tooltip = f"Sent {timestamp}"

    def on_mount(self) -> None:
        if self.message.get("role") == "assistant":
            self.add_class("assistant-message")

    def action_details(self) -> None:
        self.app.push_screen(
            MessageInfo(message=self.message, model_name=self.model_name)
        )

    @property
    def markdown(self) -> Markdown:
        return Markdown(self.message.get("content") or "")

    def render(self) -> RenderableType:
        return self.markdown

    def get_content_width(self, container: Size, viewport: Size) -> int:
        # Naive approach. Can sometimes look strange, but works well enough.
        content = self.message.get("content", "")
        return min(len(content), container.width)

    def append_chunk(self, chunk: Any):
        # If this Chatbox doesn't correspond to an OpenAI message,
        # make that connection now.
        if self.message.get("content") is None:
            # TODO - fill in the None values below
            self.message = ChatMessage(
                id=None,
                role="assistant",
                content="",
                timestamp=time.time(),
                status=None,
                end_turn=None,
                weight=None,
                metadata=None,
                recipient=None,
            )
        else:
            chunk_content = chunk["choices"][0].get("delta", {}).get("content", "")
            self.message = ChatMessage(
                id=None,
                role=self.message.get("role", "undefined"),
                content=self.message.get("content", "") + chunk_content,
                timestamp=time.time(),
                status=None,
                end_turn=None,
                weight=None,
                metadata=None,
                recipient=None,
            )
        self.refresh(layout=True)



from __future__ import annotations


import tiktoken
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll, Vertical, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Static, Tabs, ContentSwitcher, Tab

from elia_chat.models import ChatMessage
from elia_chat.time_display import format_timestamp
from elia_chat.widgets.token_analysis import TokenAnalysis


class MessageInfo(ModalScreen):
    BINDINGS = [Binding("escape", "app.pop_screen", "Close Modal")]

    def __init__(
        self,
        message: ChatMessage,
        model_name: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
        )
        self.message = message
        self.model_name = model_name

    def compose(self) -> ComposeResult:
        markdown_content = self.message.get("content", "")
        encoder = tiktoken.encoding_for_model(self.model_name)
        tokens = encoder.encode(markdown_content)

        with Vertical(id="outermost-container"):
            with Horizontal(id="message-info-header"):
                yield Tabs(
                    Tab("Markdown", id="markdown-content"),
                    Tab("Tokens", id="tokens"),
                    Tab("Metadata", id="metadata"),
                )

            with VerticalScroll(id="inner-container"):
                with ContentSwitcher(initial="markdown-content"):
                    yield Static(markdown_content, id="markdown-content")
                    yield TokenAnalysis(tokens, encoder, id="tokens")
                    yield Static("Metadata", id="metadata")

            with Horizontal(id="message-info-footer"):
                if self.model_name:
                    token_count = len(tokens)

                timestamp = self.message.get("timestamp") or 0
                timestamp_string = format_timestamp(timestamp)
                yield Static(f"Message sent at {timestamp_string}", id="timestamp")
                yield Static(f"{token_count} tokens", id="token-count")

    @on(Tabs.TabActivated)
    def tab_activated(self, event: Tabs.TabActivated) -> None:
        self.query_one(ContentSwitcher).current = event.tab.id



from sqlmodel import SQLModel

from elia_chat.database.models import engine


def create_database() -> None:
    SQLModel.metadata.create_all(engine)


if __name__ == "__main__":
    create_database()



from __future__ import annotations

import datetime
from dataclasses import dataclass

import humanize
from rich.console import RenderResult, Console, ConsoleOptions
from rich.padding import Padding
from rich.text import Text
from textual import log, on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Button, OptionList, Static
from textual.widgets.option_list import Option

from elia_chat.chats_manager import ChatsManager
from elia_chat.models import ChatData


@dataclass
class ChatListItemRenderable:
    chat: ChatData
    is_open: bool = False

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        utc_dt = datetime.datetime.utcnow()
        local_dt = utc_dt.astimezone()
        delta = local_dt - self.chat.create_time
        subtitle = humanize.naturaltime(delta)
        yield Padding(
            Text.assemble(
                (self.chat.short_preview, "" if not self.is_open else "b"),
                "\n",
                (subtitle, "dim"),
            ),
            pad=(0, 1),
            style="reverse" if self.is_open else "",
        )


class ChatListItem(Option):
    def __init__(self, chat: ChatData, is_open: bool = False) -> None:
        """
        Args:
            chat: The chat associated with this option.
            is_open: True if this is the chat that's currently open.
        """
        super().__init__(ChatListItemRenderable(chat, is_open=is_open))
        self.chat = chat
        self.is_open = is_open


class ChatList(Widget):
    COMPONENT_CLASSES = {"app-title", "app-subtitle"}

    current_chat_id: reactive[str | None] = reactive(None)

    @dataclass
    class ChatOpened(Message):
        chat: ChatData

    def compose(self) -> ComposeResult:
        with Vertical(id="cl-header-container"):
            yield Static(Text("elia", style=self.get_component_rich_style("app-title")))
            yield Static(
                Text(
                    "ChatGPT in the terminal",
                    style=self.get_component_rich_style("app-subtitle"),
                )
            )

        self.options = self.load_chat_list_items()
        option_list = OptionList(
            *self.options,
            id="cl-option-list",
        )
        yield option_list

        with Horizontal(id="cl-button-container"):
            yield Button("[Ctrl+N] New Chat", id="cl-new-chat-button")

    @on(OptionList.OptionSelected, "#cl-option-list")
    def post_chat_opened(self, event: OptionList.OptionSelected) -> None:
        assert isinstance(event.option, ChatListItem)
        chat = event.option.chat
        self.current_chat_id = chat.id
        self.reload_and_refresh()
        self.post_message(ChatList.ChatOpened(chat=chat))

    def on_focus(self) -> None:
        log.debug("Sidebar focused")
        self.query_one("#cl-option-list", OptionList).focus()

    def reload_and_refresh(self, new_highlighted: int = -1) -> None:
        """Reload the chats and refresh the widget. Can be used to
        update the ordering/previews/titles etc contained in the list.

        Args:
            new_highlighted: The index to highlight after refresh.
        """
        self.options = self.load_chat_list_items()
        option_list = self.query_one(OptionList)
        old_highlighted = option_list.highlighted
        option_list.clear_options()
        option_list.add_options(self.options)
        if new_highlighted > -1:
            option_list.highlighted = new_highlighted
        else:
            option_list.highlighted = old_highlighted

    def load_chat_list_items(self) -> list[ChatListItem]:
        chats = self.load_chats()
        return [
            ChatListItem(chat, is_open=self.current_chat_id == chat.id)
            for chat in chats
        ]

    def load_chats(self) -> list[ChatData]:
        all_chats = ChatsManager.all_chats()
        return all_chats

    def create_chat(self, chat_data: ChatData) -> None:
        new_chat_list_item = ChatListItem(chat_data, is_open=True)
        log.debug(f"Creating new chat {new_chat_list_item!r}")

        option_list = self.query_one(OptionList)
        self.options = [
            new_chat_list_item,
            *self.options,
        ]
        option_list.clear_options()
        option_list.add_options(self.options)
        option_list.highlighted = 0
        self.refresh()



from datetime import datetime


from elia_chat.database.models import ChatDao, MessageDao
from elia_chat.models import ChatData, ChatMessage


def chat_data_to_chat_dao(chat_data: ChatData) -> ChatDao:
    return ChatDao(
        model=chat_data.model_name,
        started_at=datetime.fromtimestamp(chat_data.create_timestamp or 0),
    )


def chat_message_to_message_dao(chat_message: ChatMessage) -> MessageDao:
    return MessageDao(
        role=chat_message["role"],
        content=chat_message["content"],
        timestamp=datetime.fromtimestamp(chat_message["timestamp"]),
        status=chat_message["status"],
        end_turn=chat_message["end_turn"],
        weight=chat_message["weight"],
        meta=chat_message["metadata"],
        recipient=chat_message["recipient"],
    )


def chat_dao_to_chat_data(chat_dao: ChatDao) -> ChatData:
    return ChatData(
        id=str(chat_dao.id),
        title=chat_dao.title,
        model_name=chat_dao.model,
        create_timestamp=chat_dao.started_at.timestamp()
        if chat_dao.started_at
        else None,
        messages=[
            message_dao_to_chat_message(message) for message in chat_dao.messages
        ],
    )


def message_dao_to_chat_message(message_dao: MessageDao) -> ChatMessage:
    return ChatMessage(
        id=str(message_dao.id),
        role=message_dao.role,
        content=message_dao.content,
        timestamp=message_dao.timestamp.timestamp() if message_dao.timestamp else 0,
        status=message_dao.status,
        end_turn=message_dao.end_turn,
        weight=message_dao.weight,
        metadata=message_dao.meta,
        recipient=message_dao.recipient,
    )



from __future__ import annotations

import pathlib
from datetime import datetime
from typing import Any

from sqlalchemy import Column, DateTime, func, JSON, desc
from sqlalchemy.orm import selectinload
from sqlmodel import SQLModel, Field, create_engine, Session, select, Relationship


class MessageDao(SQLModel, table=True):
    __tablename__ = "message"

    id: int = Field(default=None, primary_key=True)
    chat_id: int | None = Field(foreign_key="chat.id")
    chat: ChatDao = Relationship(back_populates="messages")
    role: str
    content: str
    timestamp: datetime | None = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now())
    )
    status: str | None
    end_turn: bool | None
    weight: float | None
    meta: dict = Field(sa_column=Column(JSON), default={})
    recipient: str | None


class ChatDao(SQLModel, table=True):
    __tablename__ = "chat"

    id: int = Field(default=None, primary_key=True)
    model: str | None
    title: str | None
    started_at: datetime | None = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now())
    )
    messages: list[MessageDao] = Relationship(back_populates="chat")

    @staticmethod
    def all() -> list[ChatDao]:
        with Session(engine) as session:
            # Create a subquery that finds the maximum
            # (most recent) timestamp for each chat.
            max_timestamp: Any = func.max(MessageDao.timestamp).label("max_timestamp")
            subquery = (
                select(MessageDao.chat_id, max_timestamp)
                .group_by(MessageDao.chat_id)
                .alias("subquery")
            )

            statement = (
                select(ChatDao)
                .join(subquery, subquery.c.chat_id == ChatDao.id)
                .order_by(desc(subquery.c.max_timestamp))
                .options(selectinload(ChatDao.messages))
            )
            results = session.exec(statement)
            return list(results)

    @staticmethod
    def from_id(chat_id: str) -> ChatDao:
        with Session(engine) as session:
            statement = (
                select(ChatDao)
                .where(ChatDao.id == int(chat_id))
                .options(selectinload(ChatDao.messages))
            )
            result = session.exec(statement).one()
            return result


_this_dir = pathlib.Path(__file__).resolve().parent
sqlite_file_name = _this_dir / "elia.sqlite"
sqlite_url = f"sqlite:///{sqlite_file_name}"
engine = create_engine(sqlite_url)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TypedDict


class ChatMessage(TypedDict):
    id: str | None
    role: str
    content: str
    timestamp: float
    status: str | None
    end_turn: bool | None
    weight: float | None
    metadata: dict | None
    recipient: str | None


@dataclass
class ChatData:
    id: str | None
    model_name: str | None
    title: str | None
    create_timestamp: float | None
    messages: list[ChatMessage]

    @property
    def short_preview(self) -> str:
        first_user_message = self.first_user_message
        if first_user_message is None:
            return "Empty chat..."
        return first_user_message.get("content", "")[:24] + "..."

    @property
    def first_user_message(self) -> ChatMessage | None:
        return self.messages[1] if len(self.messages) > 1 else None

    @property
    def non_system_messages(self) -> list[ChatMessage]:
        return self.messages[1:]

    @property
    def create_time(self) -> datetime:
        return datetime.fromtimestamp(self.create_timestamp or 0).astimezone()

    @property
    def update_time(self) -> datetime:
        return datetime.fromtimestamp(
            self.messages[-1].get("timestamp", 0) if self.messages else 0
        )



"""
Elia CLI
"""

import pathlib

import click

from elia_chat.app import app
from elia_chat.database.create_database import create_database
from elia_chat.database.import_chatgpt import import_chatgpt_data
from elia_chat.database.models import sqlite_file_name


@click.group(invoke_without_command=True)
@click.pass_context
def cli(context: click.Context) -> None:
    """
    Elia: A terminal ChatGPT client built with Textual
    """
    # Run the app if no subcommand is provided
    if context.invoked_subcommand is None:
        # Create the database if it doesn't exist
        if sqlite_file_name.exists() is False:
            create_database()
        app.run()


@cli.command()
def reset() -> None:
    """
    Reset the database

    This command will delete the database file and recreate it.
    Previously saved conversations and data will be lost.
    """
    sqlite_file_name.unlink(missing_ok=True)
    create_database()
    click.echo(f"♻️  Database reset @ {sqlite_file_name}")


@cli.command("import")
@click.argument(
    "file",
    type=click.Path(
        exists=True, dir_okay=False, path_type=pathlib.Path, resolve_path=True
    ),
)
def import_file_to_db(file: pathlib.Path) -> None:
    """
    Import ChatGPT Conversations

    This command will import the ChatGPT conversations from a local
    JSON file into the database.
    """
    import_chatgpt_data(file=file)
    click.echo(f"✅  ChatGPT data imported into database {file}")


if __name__ == "__main__":
    cli()



from __future__ import annotations

from textual.reactive import reactive
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import Static


class ChatHeader(Widget):
    title = reactive("Untitled Chat", init=False)
    model_name = reactive("", init=False)

    def watch_model_name(self, new_model_name: str | None) -> None:
        if new_model_name is not None:
            model_static = self.query_one("#model-static", Static)
            model_static.update(new_model_name)

    def watch_title(self, new_title: str | None) -> None:
        if new_title is not None:
            title_static = self.query_one("#title-static", Static)
            title_static.update(new_title)

    def compose(self) -> ComposeResult:
        yield Static(self.title, id="title-static")
        yield Static(self.model_name, id="model-static")
